package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experimenter_gui.util.ui.MacWithDotsBinding;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.ReadOnlyTextFieldTableCell;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.DiskImage;
import be.iminds.ilabt.jfed.rspec.model.RspecInterface;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecInterface;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ExternalSshTerminal;
import be.iminds.ilabt.jfed.ui.javafx.FXMLUtil;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.util.common.OSDetector;
import javafx.beans.binding.Bindings;
import javafx.beans.binding.DoubleBinding;
import javafx.beans.property.SimpleStringProperty;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.layout.GridPane;
import javafx.scene.paint.Color;
import javafx.util.Pair;
import javafx.util.StringConverter;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;

/**
 * A dialog that shows some simple information about a node: authentication-type, hostname, login and port
 * <p>
 * User: twalcari
 * Date: 1/17/14
 * Time: 10:33 AM
 */
public class NodePropertiesDialog extends GridPane {
    private static final Logger LOG = LoggerFactory.getLogger(NodePropertiesDialog.class);

    @Nonnull private final Slice slice;
    @Nonnull private final FXRspecNode node;
    @Nonnull private final BasicStringRspec.LoginService bestLoginService;
    @Nonnull private final List<BasicStringRspec.LoginService> loginServices;
    @Nonnull private final ExternalSshTerminal externalSshTerminal;

    @FXML
    private ComboBox<BasicStringRspec.LoginService> loginCombobox;

    @FXML
    private Label authenticationTypeLabel;
    @FXML
    private TextField hostnameTextField;
    @FXML
    private TextField usernameTextField;
    @FXML
    private TextField portTextField;
    @FXML
    private TextArea cliSshTextArea;

    @FXML
    private TextField diskImageTextField;

    @FXML
    private TableView<Pair<FXRspecInterface, FXRspecInterface.IpAddress>> interfacesTableView;

    @FXML
    private TableColumn<Pair<FXRspecInterface, FXRspecInterface.IpAddress>, String> ifaceIdTableColumn;
    @FXML
    private TableColumn<Pair<FXRspecInterface, FXRspecInterface.IpAddress>, String> ifaceIpTableColumn;
    @FXML
    private TableColumn<Pair<FXRspecInterface, FXRspecInterface.IpAddress>, String> ifaceNetmaskTableColumn;
    @FXML
    private TableColumn<Pair<FXRspecInterface, FXRspecInterface.IpAddress>, String> ifaceMacTableColumn;
    @FXML
    private TableColumn<Pair<FXRspecInterface, FXRspecInterface.IpAddress>, String> ifaceLinkTableColumn;


    private final Dialog<ButtonType> dialog;

    NodePropertiesDialog(@Nonnull Slice slice,
                         @Nonnull FXRspecNode node,
                         @Nonnull BasicStringRspec.LoginService bestLoginService,
                         @Nonnull List<BasicStringRspec.LoginService> loginServices,
                         @Nonnull ExternalSshTerminal externalSshTerminal) {
        assert !loginServices.isEmpty();
        this.slice = slice;
        this.node = node;
        this.bestLoginService = bestLoginService;
        this.loginServices = loginServices;
        this.externalSshTerminal = externalSshTerminal;

        //initialize the Scene
        FXMLUtil.injectFXML(this);

        DoubleBinding tableWidth = interfacesTableView.widthProperty().subtract(4);

        ifaceIdTableColumn.prefWidthProperty().bind(tableWidth.multiply(0.125));
        ifaceIpTableColumn.prefWidthProperty().bind(tableWidth.multiply(0.25));
        ifaceNetmaskTableColumn.prefWidthProperty().bind(tableWidth.multiply(0.25));
        ifaceMacTableColumn.prefWidthProperty().bind(tableWidth.multiply(0.25));
        ifaceLinkTableColumn.prefWidthProperty().bind(tableWidth.multiply(0.125));

        ifaceIdTableColumn.setCellValueFactory(param ->
                param.getValue().getKey().clientIdProperty()
        );
        ifaceMacTableColumn.setCellValueFactory(param -> new MacWithDotsBinding(
                param.getValue().getKey().macAddressProperty()
        ));
        ifaceIpTableColumn.setCellValueFactory(param -> new SimpleStringProperty(
                param.getValue() != null &&
                        param.getValue().getValue() != null &&
                        param.getValue().getValue().getAddress() != null ?
                        param.getValue().getValue().getAddress() :
                        ""));
        ifaceNetmaskTableColumn.setCellValueFactory(param -> new SimpleStringProperty(
                param.getValue() != null  &&
                        param.getValue().getValue() != null &&
                        param.getValue().getValue().getNetmask() != null?
                        param.getValue().getValue().getNetmask() :
                        ""));
        ifaceLinkTableColumn.setCellValueFactory(param -> {
            try {
                return new SimpleStringProperty(
                        param.getValue() != null
                                && param.getValue().getKey() != null
                                && param.getValue().getKey().getLink() != null
                                && param.getValue().getKey().getLink().getClientId() != null ?
                                param.getValue().getKey().getLink().getClientId() : ""
                );
            } catch (Exception e) {
                LOG.error("Exception showing link info in table", e);
                return new SimpleStringProperty("**ERROR**");
            }
        });


        interfacesTableView.prefHeightProperty().bind(Bindings.size(interfacesTableView.getItems()).multiply(28).add(25 /*header */ + 12 /*safety margin */));

        ifaceIpTableColumn.setCellFactory(ReadOnlyTextFieldTableCell.forTableColumn());
        ifaceNetmaskTableColumn.setCellFactory(ReadOnlyTextFieldTableCell.forTableColumn());
        ifaceMacTableColumn.setCellFactory(ReadOnlyTextFieldTableCell.forTableColumn());

        if (this.loginServices.size() > 1) {
            loginCombobox.setConverter(new StringConverter<BasicStringRspec.LoginService>() {
                @Override
                public String toString(BasicStringRspec.LoginService ls) {
                    return ls.toDescription(true, false);
                }

                @Override
                public BasicStringRspec.LoginService fromString(String string) {
                    throw new RuntimeException("Cannot convert from String to LoginService");
                }
            });
            loginCombobox.getItems().setAll(this.loginServices);
            loginCombobox.getSelectionModel().select(this.bestLoginService);
            showLoginInfo(this.bestLoginService);
            loginCombobox.getSelectionModel().selectedItemProperty().addListener(
                    (observable, oldValue, newValue) -> {
                        if (newValue != null) showLoginInfo(newValue);
                    });
        } else {
            showLoginInfo(this.bestLoginService);
            loginCombobox.setVisible(false);
        }

        if (node.getSliverType() != null) {
            if (node.getSliverType().getDiskImages().isEmpty()) {
                diskImageTextField.setVisible(true);
                diskImageTextField.setManaged(true);
                diskImageTextField.setText("<Testbed Default>");
                diskImageTextField.setDisable(true);
            } else {
                showDiskImageInfo(node.getSliverType().getDiskImages().get(0));
            }
        } else {
            diskImageTextField.setVisible(false);
            diskImageTextField.setManaged(false);
        }

        List<Pair<FXRspecInterface, FXRspecInterface.IpAddress>> ifaceWithAddressCombos =
                new ArrayList<>();

        for (FXRspecInterface iface : node.getInterfaces()) {
            if (!iface.getIpAddresses().isEmpty()) {
                for (RspecInterface.IpAddress ipAddress : iface.getIpAddresses()) {
                    ifaceWithAddressCombos.add(new Pair<>(iface, ipAddress));
                }
            } else {
                ifaceWithAddressCombos.add(new Pair<>(iface, null));
            }
        }

        if (ifaceWithAddressCombos.isEmpty()) {
            interfacesTableView.setVisible(false);
            interfacesTableView.setManaged(false);
        } else {
            interfacesTableView.getItems().addAll(ifaceWithAddressCombos);
        }


        DialogPane dialogPane = new DialogPane();
        dialogPane.setContent(this);
        dialogPane.setHeaderText("Properties of " + node.getClientId());
        dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.INFO, Color.BLACK));
        dialogPane.getButtonTypes().setAll(ButtonType.CLOSE);

        //initialize dialog
        dialog = new javafx.scene.control.Dialog<>();
        dialog.setDialogPane(dialogPane);
        dialog.setTitle(dialogPane.getHeaderText());
        dialog.setHeaderText(dialog.getTitle());
        dialog.setResizable(true);


    }

    private void showDiskImageInfo(@Nonnull DiskImage diskImage) {
        if (diskImage.getName() != null) {
            diskImageTextField.setVisible(true);
            diskImageTextField.setManaged(true);
            diskImageTextField.setDisable(false);
            diskImageTextField.setText(diskImage.getName());
            return;
        }
        if (diskImage.getUrl() != null) {
            diskImageTextField.setVisible(true);
            diskImageTextField.setManaged(true);
            diskImageTextField.setDisable(false);
            diskImageTextField.setText(diskImage.getUrl());
            return;
        }
        diskImageTextField.setVisible(false);
        diskImageTextField.setManaged(false);
    }

    private void showLoginInfo(BasicStringRspec.LoginService loginService) {
        authenticationTypeLabel.setText(loginService.getAuthentication());
        hostnameTextField.setText(loginService.getHostname());
        portTextField.setText(loginService.getPort() + "");
        usernameTextField.setText(loginService.getUsername());

        if (OSDetector.getUnixLikeOsFilter().includedOs(OSDetector.os)) {
            try {
                String cli = getCliSshCommand(loginService);
                cliSshTextArea.setText(cli);
            } catch (Exception e) {
                LOG.error("Something went wrong generating the SSH command line.");
                cliSshTextArea.setText("(error generating ssh command)");
            }
        } else {
            cliSshTextArea.setVisible(false);
        }
    }

    @Nonnull
    private String getCliSshCommand(@Nonnull BasicStringRspec.LoginService loginService) {
        return externalSshTerminal.getCliSshCommand(
                loginService.getUsername(),
                loginService.getHostname(),
                loginService.getPort(),
                loginService.getSshProxy()
        );
    }


    public void showDialog() {
        dialog.show();
    }

}
